// SPDX-License-Identifier: MIT
pragma solidity ^0.8.4;

This file exists to show that a brownie-style dependency can contain a contract that does not compile.
But the dependency is still usable (provided the project only uses working contract types)
